from cytoolz.curried import curry
from collections import deque
from collections.abc import Iterable

from .base import FileSink
from govcf.variant_files import ensure_collection
from genomoncology.parse import DocType

import related


@related.immutable()
class DummyVariantRecord(object):
    chrom = related.StringField()
    pos = related.IntegerField()
    ref = related.StringField()
    alt = related.StringField()
    info = related.ChildField(dict, default=dict)

    CHR_FIELDS = ["chr", "c"]
    POS_FIELDS = ["pos", "s", "start", "hg19_pos", "hg37_pos", "hg38_pos"]
    REF_FIELDS = ["ref", "r"]
    ALT_FIELDS = ["alt", "a"]
    INFO_IGNORE = {"__type__", "build"}

    def __str__(self):
        return (
            "\t".join(
                [
                    self.chrom,
                    str(self.pos),
                    ".",
                    self.ref,
                    self.alt,
                    ".",
                    ".",
                    self.info_str,
                ]
            )
            + "\n"
        )

    @property
    def info_str(self):
        return ";".join(self.info_nvp())

    def info_nvp(self):
        for (name, value) in self.info.items():
            if name == "info":
                ivalues = value[0].items()
                for (iname, ivalue) in ivalues:
                    yield f"{iname}={ivalue}"

            elif name not in self.INFO_IGNORE:
                value_str = ",".join(map(str, value))
                yield f"{name}={value_str}"

    @classmethod
    def pop_first(cls, unit: dict, field_names: list):
        for field_name in field_names:
            value = unit.pop(field_name, None)
            if value is not None:
                return value

    @classmethod
    def create_from_unit(cls, unit: dict):
        unit = unit.copy()
        chrom = cls.pop_first(unit, cls.CHR_FIELDS)
        pos = cls.pop_first(unit, cls.POS_FIELDS)
        ref = cls.pop_first(unit, cls.REF_FIELDS)
        alt = cls.pop_first(unit, cls.ALT_FIELDS)

        info = {}
        for (name, value) in unit.items():
            info[name] = ensure_collection(value)

        return DummyVariantRecord(chrom, pos, ref, alt, info)


@curry
class HistoricalVcfFileSink(FileSink):
    def __init__(self, filename, header_file_path=None):
        super().__init__(filename, insert_newlines=False)
        self.recent_records_seen = deque(maxlen=100)
        self.count = 0
        self.header_file_path = header_file_path
        # map from original record to a dict containing:
        #   - the total number of alts for this record
        #   - the number of the alts already encountered
        #   - the updated record object (with annotations
        #     from its alts in the INFO column)
        self.multi_alt_lines = {}

    def handle_key(self, key, alt_index=None):
        return "%s__%d" % (key, alt_index) if alt_index else key

    def handle_value(self, values):
        if values is None:
            return ''
        if isinstance(values, Iterable) and not isinstance(values, (str, bytes)):
            return "|".join(
                [
                    ascii(value).replace(" ", "_").replace("'", "")
                    for value in values
                ]
            )
        return ascii(values).replace(" ", "_").replace("'", "")

    def handle_gene_annotations(
        self, info_dict, gene_annotations, alt_index=None
    ):
        if isinstance(gene_annotations, list):
            self.format_gene_annotations(
                gene_annotations, info_dict, alt_index
            )
        else:
            for gene_annotation_key, gene_anns in gene_annotations.items():
                self.format_gene_annotations(gene_anns, info_dict, alt_index)

    def format_gene_annotations(self, gene_anns, info_dict, alt_index=None):
        for gene_ann in gene_anns:
            gene_name = gene_ann.pop("gene", "")
            if gene_ann:
                for k, v in gene_ann.items():
                    key = self.handle_key("%s__%s" % (k, gene_name), alt_index)
                    value = self.handle_value(v)
                    if value:
                        info_dict[key] = value

    def add_annotations_to_info(self, existing_info, unit, alt_index=None):
        # handle existing info
        if not existing_info or existing_info == ".":
            existing_info = ""

        annotations = unit.get("annotations", {})
        info_dict = {}

        # add annotation key-value pairs to info_dict
        for key, value in annotations.items():
            if key == "gene_annotations":
                self.handle_gene_annotations(info_dict, value, alt_index)
                continue
            value = self.handle_value(annotations.get(key))
            key = self.handle_key(key, alt_index)
            if value:
                info_dict[key] = value

        # merge all info values into a semicolon-separated string
        info_strings = [f"{key}={value}" for key, value in info_dict.items()]

        # add semicolon to end of existing info if there isn't one already
        if info_strings and existing_info and existing_info[-1] != ";":
            existing_info += ";"

        updated_info = existing_info + ";".join(info_strings)
        if not updated_info:
            updated_info = "."

        return updated_info

    def handle_historical(self, record, unit, alt_index=None):
        columns = record.split("\t")
        if len(columns) >= 10 and not columns[0].startswith("#"):
            # add annotations from unit to record_info
            record_info = columns[7]
            columns[7] = self.add_annotations_to_info(
                record_info, unit, alt_index
            )
        return "\t".join(columns)

    def get_record_alt(self, record):
        columns = record.split("\t")
        if len(columns) >= 10 and not columns[0].startswith("#"):
            return columns[4].split(",")
        return []

    def handle_multi_alt(self, record, unit, record_alt):
        # figure out what alt_index this is
        unit_alt = unit.get("alt", "")
        alt_index = record_alt.index(unit_alt)
        if alt_index < 0:
            return record

        # see if this record has already been encountered!
        multi_alt_record = self.multi_alt_lines.get(record, {})
        if not multi_alt_record:
            # set up default value
            multi_alt_record = {
                "total_num_alts": len(record_alt),
                "num_alts_encountered": 0,
                "updated_record": record,
            }
        multi_alt_record["num_alts_encountered"] = (
            multi_alt_record.get("num_alts_encountered") + 1
        )
        multi_alt_record["updated_record"] = self.handle_historical(
            multi_alt_record.get("updated_record"), unit, alt_index + 1
        )
        self.multi_alt_lines[record] = multi_alt_record

        num_encountered = multi_alt_record.get("num_alts_encountered")
        total_num_alts = multi_alt_record.get("total_num_alts")
        if num_encountered == total_num_alts:
            # we have encountered all of the alts for this record!
            # No need to keep it in the map.
            _ = self.multi_alt_lines.pop(record)
            return multi_alt_record.get("updated_record")
        else:
            return None

    def get_header(self):
        if self.header_file_path is not None:
            with open(self.header_file_path) as header_file:
                header_file_lines = header_file.readlines()
                return "".join(header_file_lines)
        return DEFAULT_VCF_HEADER

    def get_record_str(self, unit) -> str:
        record = unit.get("__record__")
        if record is None:
            if DocType.HEADER.is_a(unit):
                record = self.get_header()
            else:
                record = DummyVariantRecord.create_from_unit(unit)
        # even if there IS a record, if this is a header
        # and we have a header override, use the override
        if DocType.HEADER.is_a(unit) and self.header_file_path is not None:
            record = self.get_header()
        record = str(record)
        return record

    def convert(self, unit):
        self.count += 1
        record = self.get_record_str(unit)

        # handle this record if is is a multi-alt record
        record_alt = self.get_record_alt(record)
        is_multi_alt = len(record_alt) > 1
        if is_multi_alt:
            # if multi_alt_record is None, this multi-alt record is not yet
            # ready to be written to the VCF because we have not yet seen
            # all of its constituent alts. But if we HAVE seen them all,
            # multi_alt_record will be populated with the record string
            # (with the annotations from all the constituent alts in
            # the INFO column)
            multi_alt_record = self.handle_multi_alt(record, unit, record_alt)
            if multi_alt_record:
                return multi_alt_record

        if record not in self.recent_records_seen and not is_multi_alt:
            self.recent_records_seen.append(record)
            # add the annotations to the INFO column.
            record = self.handle_historical(record, unit)
            return record


@curry
class VcfFileSink(FileSink):
    def __init__(self, filename, header_file_path=None):
        super().__init__(filename, insert_newlines=False)
        self.recent_records_seen = deque(maxlen=100)
        self.count = 0
        self.header_file_path = header_file_path

    def get_header(self):
        if self.header_file_path is not None:
            with open(self.header_file_path) as header_file:
                header_file_lines = header_file.readlines()
                return "".join(header_file_lines)
        return DEFAULT_VCF_HEADER

    def get_record_str(self, unit) -> str:
        record = unit.get("__record__")
        if record is None:
            if DocType.HEADER.is_a(unit):
                record = self.get_header()
            else:
                record = DummyVariantRecord.create_from_unit(unit)
        # even if there IS a record, if this is a header
        # and we have a header override, use the override
        if DocType.HEADER.is_a(unit) and self.header_file_path is not None:
            record = self.get_header()
        record = str(record)
        return record

    def convert(self, unit):
        self.count += 1
        record = self.get_record_str(unit)

        if record not in self.recent_records_seen:
            self.recent_records_seen.append(record)
            return record


DEFAULT_VCF_HEADER = """##fileformat=VCFv4.2
##contig=<ID=1,length=249250621>
##contig=<ID=2,length=243199373>
##contig=<ID=3,length=198022430>
##contig=<ID=4,length=191154276>
##contig=<ID=5,length=180915260>
##contig=<ID=6,length=171115067>
##contig=<ID=7,length=159138663>
##contig=<ID=8,length=146364022>
##contig=<ID=9,length=141213431>
##contig=<ID=10,length=135534747>
##contig=<ID=11,length=135006516>
##contig=<ID=12,length=133851895>
##contig=<ID=13,length=115169878>
##contig=<ID=14,length=107349540>
##contig=<ID=15,length=102531392>
##contig=<ID=16,length=90354753>
##contig=<ID=17,length=81195210>
##contig=<ID=18,length=78077248>
##contig=<ID=19,length=59128983>
##contig=<ID=20,length=63025520>
##contig=<ID=21,length=48129895>
##contig=<ID=22,length=51304566>
##contig=<ID=X,length=155270560>
##contig=<ID=Y,length=59373566>
##contig=<ID=MT,length=16569>
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
"""



from genomoncology import kms
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


def convert_ac_to_vaf(x, field):
    return (
        float(x[field][x["orig_alts__mstring"].index(x["alt"]) + 1])
        / sum([float(i) for i in x[field]])
    ) * 100


NAME_MAPPING = {
    # hgvs
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    "build": "build",
    "orig_alts__mstring": "__record__",
    # info
    "EA__float": "info.EA_AC",
    "AA__float": "info.AA_AC",
    "All__float": "info.TAC",
}

register(
    input_type=DocType.CALL,
    output_type=DocType.EVS,
    transformer=compose(
        lambda x: assoc(x, "EA__float", convert_ac_to_vaf(x, "EA__float")),
        lambda x: assoc(x, "AA__float", convert_ac_to_vaf(x, "AA__float")),
        lambda x: assoc(x, "All__float", convert_ac_to_vaf(x, "All__float")),
        lambda x: assoc(x, "orig_alts__mstring", x["orig_alts__mstring"].alts),
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        lambda x: assoc(x, __TYPE__, DocType.EVS.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.CALL,
    output_type=DocType.EVS,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.EVS.value)),
    is_header=True,
)



from cytoolz.curried import assoc, compose, merge

from genomoncology import kms
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import (
    register,
    name_mapping,
    rename,
    split_value,
)

SPLIT_CHARS = ":;"


def set_chr_start(value):
    if value.get("build", "").endswith("38"):
        value = rename("#chr", "chr", value)
        value = rename("pos(1-based)", "start", value)
    else:
        value = rename("hg19_chr", "chr", value)
        value = rename("hg19_pos(1-based)", "start", value)

    return value


def get_largest_value(value):
    return max(value) if value else None


def additional_clean_up(value):
    genes = split_value(value.get("gene", []), split_chars=SPLIT_CHARS)

    updates = dict(gene=genes)

    return merge(value, updates)


def filter_out_empty_chr_start(value):
    if value.get("chr") is None and value.get("start") is None:
        return None
    return value


NAME_MAPPING = {
    # critical
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    "build": "build",
    "gene": "genename",
    "grch38_csra": "grch38_csra",
    # baylor
    "1000Gp3_AC__int": "1000Gp3_AC",
    "1000Gp3_AF__float": "1000Gp3_AF",
    "1000Gp3_AFR_AC__int": "1000Gp3_AFR_AC",
    "1000Gp3_AFR_AF__float": "1000Gp3_AFR_AF",
    "1000Gp3_AMR_AC__int": "1000Gp3_AMR_AC",
    "1000Gp3_AMR_AF__float": "1000Gp3_AMR_AF",
    "1000Gp3_EAS_AC__int": "1000Gp3_EAS_AC",
    "1000Gp3_EAS_AF__float": "1000Gp3_EAS_AF",
    "1000Gp3_EUR_AC__int": "1000Gp3_EUR_AC",
    "1000Gp3_EUR_AF__float": "1000Gp3_EUR_AF",
    "1000Gp3_SAS_AC__int": "1000Gp3_SAS_AC",
    "1000Gp3_SAS_AF__float": "1000Gp3_SAS_AF",
    "cds_strand__mstring": "cds_strand",
    "MutationTaster_AAE__mstring": "MutationTaster_AAE",
    "MutationTaster_converted_rankscore__float": "MutationTaster_converted_rankscore",  # noqa E501
    "MutationTaster_model__mstring": "MutationTaster_model",
    "MutationTaster_pred__mstring": "MutationTaster_pred",
    "MutationTaster_score__mfloat": "MutationTaster_score",
    "SIFT_converted_rankscore__mfloat": "SIFT_converted_rankscore",
    "SIFT_pred__mstring": "SIFT_pred",
    "SIFT_score__mfloat": "SIFT_score",
    # not used (yet)
    "PROVEAN_converted_rankscore__float": "PROVEAN_converted_rankscore",
    "PROVEAN_pred__mstring": "PROVEAN_pred",
    "PROVEAN_score__mfloat": "PROVEAN_score",
    "codonpos__mint": "codonpos",
    "codon_degeneracy__mstring": "codon_degeneracy",
    "Ancestral_allele__string": "Ancestral_allele",
    "AltaiNeandertal__string": "AltaiNeandertal",
    "Denisova__string": "Denisova",
    "Ensembl_geneid__mstring": "Ensembl_geneid",
    "Ensembl_transcriptid__mstring": "Ensembl_transcriptid",
    "Ensembl_proteinid__mstring": "Ensembl_proteinid",
    "LRT_score__float": "LRT_score",
    "LRT_converted_rankscore__float": "LRT_converted_rankscore",
    "LRT_pred__string": "LRT_pred",
    "LRT_Omega__float": "LRT_Omega",
    "MutationAssessor_score__mstring": "MutationAssessor_score",
    "MutationAssessor_pred__mstring": "MutationAssessor_pred",
    "FATHMM_score__mfloat": "FATHMM_score",
    "FATHMM_converted_rankscore__float": "FATHMM_converted_rankscore",
    "FATHMM_pred__mstring": "FATHMM_pred",
    "MetaSVM_score__float": "MetaSVM_score",
    "MetaSVM_rankscore__float": "MetaSVM_rankscore",
    "MetaSVM_pred__string": "MetaSVM_pred",
    "MetaLR_score__float": "MetaLR_score",
    "MetaLR_rankscore__float": "MetaLR_rankscore",
    "MetaLR_pred__string": "MetaLR_pred",
    "Reliability_index__float": "Reliability_index",
    "M-CAP_score__float": "M-CAP_score",
    "M-CAP_rankscore__float": "M-CAP_rankscore",
    "M-CAP_pred__string": "M-CAP_pred",
    "MutPred_score__string": "MutPred_score",
    "MutPred_rankscore__string": "MutPred_rankscore",
    "MutPred_protID__string": "MutPred_protID",
    "MutPred_AAchange__string": "MutPred_AAchange",
    "MutPred_Top5features__mstring": "MutPred_Top5features",
    "fathmm-MKL_coding_score__float": "fathmm-MKL_coding_score",
    "fathmm-MKL_coding_rankscore__float": "fathmm-MKL_coding_rankscore",
    "fathmm-MKL_coding_pred__string": "fathmm-MKL_coding_pred",
    "fathmm-MKL_coding_group__string": "fathmm-MKL_coding_group",
    "Eigen-raw__float": "Eigen-raw_coding",
    "Eigen-phred__float": "Eigen-phred_coding",
    "Eigen-raw_rankscore__float": "Eigen-raw_coding_rankscore",
    "Eigen-PC-raw__float": "Eigen-PC-raw_coding",
    "Eigen-PC-phred__float": "Eigen-PC-phred_coding",
    "Eigen-PC-raw_rankscore__float": "Eigen-PC-raw_coding_rankscore",
    "integrated_fitCons_score__float": "integrated_fitCons_score",
    "integrated_confidence_value__float": "integrated_confidence_value",
    "GM12878_fitCons_score__float": "GM12878_fitCons_score",
    "GM12878_fitCons_rankscore__float": "GM12878_fitCons_rankscore",
    "GM12878_confidence_value__float": "GM12878_confidence_value",
    "H1-hESC_fitCons_score__float": "H1-hESC_fitCons_score",
    "H1-hESC_fitCons_rankscore__float": "H1-hESC_fitCons_rankscore",
    "H1-hESC_confidence_value__float": "H1-hESC_confidence_value",
    "HUVEC_fitCons_score__float": "HUVEC_fitCons_score",
    "HUVEC_fitCons_rankscore__float": "HUVEC_fitCons_rankscore",
    "HUVEC_confidence_value__float": "HUVEC_confidence_value",
    "GERP++_NR__float": "GERP++_NR",
    "GERP++_RS__float": "GERP++_RS",
    "GERP++_RS_rankscore__float": "GERP++_RS_rankscore",
    "phyloP100way_vertebrate__float": "phyloP100way_vertebrate",
    "phyloP100way_vertebrate_rankscore__float": "phyloP100way_vertebrate_rankscore",  # noqa E501
    "phyloP30way_mammalian__float": "phyloP470way_mammalian",
    "phyloP30way_mammalian_rankscore__float": "phyloP470way_mammalian_rankscore",  # noqa E501
    "phastCons100way_vertebrate__float": "phastCons100way_vertebrate",
    "phastCons100way_vertebrate_rankscore__float": "phastCons100way_vertebrate_rankscore",  # noqa E501
    "phastCons30way_mammalian__float": "phastCons470way_mammalian",
    "phastCons30way_mammalian_rankscore__float": "phastCons470way_mammalian_rankscore",  # noqa E501
    "SiPhy_29way_pi__mfloat": "SiPhy_29way_pi",
    "SiPhy_29way_logOdds__float": "SiPhy_29way_logOdds",
    "SiPhy_29way_logOdds_rankscore__float": "SiPhy_29way_logOdds_rankscore",
    "TWINSUK_AC__int": "TWINSUK_AC",
    "TWINSUK_AF__float": "TWINSUK_AF",
    "ALSPAC_AC__int": "ALSPAC_AC",
    "ALSPAC_AF__float": "ALSPAC_AF",
    "Interpro_domain__mstring": "Interpro_domain",
    "GTEx_V7p_gene__string": "GTEx_V8_gene",
    "GTEx_V7p_tissue__string": "GTEx_V8_tissue",
    "MutationAssessor_score_rankscore__string": "MutationAssessor_rankscore",  # noqa E501
    "HGVSc_ANNOVAR__mstring": "HGVSc_ANNOVAR",
    "HGVSp_ANNOVAR__mstring": "HGVSp_ANNOVAR",
    "HGVSc_snpEff__mstring": "HGVSc_snpEff",
    "HGVSp_snpEff__mstring": "HGVSp_snpEff",
    "HGVSc_VEP__mstring": "HGVSc_VEP",
    "HGVSp_VEP__mstring": "HGVSp_VEP",
    "Polyphen2_HDIV_pred__mstring": "Polyphen2_HDIV_pred",
    "Polyphen2_HDIV_rankscore__float": "Polyphen2_HDIV_rankscore",
    "Polyphen2_HDIV_score__mfloat": "Polyphen2_HDIV_score",
    "Polyphen2_HVAR_pred__mstring": "Polyphen2_HVAR_pred",
    "Polyphen2_HVAR_rankscore__float": "Polyphen2_HVAR_rankscore",
    "Polyphen2_HVAR_score__mfloat": "Polyphen2_HVAR_score",
    "SIFT4G_converted_rankscore__float": "SIFT4G_converted_rankscore",
    "SIFT4G_pred__mstring": "SIFT4G_pred",
    "SIFT4G_score__mfloat": "SIFT4G_score",
    # CADD
    "CADD_raw__float": "CADD_raw",
    "CADD_raw_rankscore__float": "CADD_raw_rankscore",
    "CADD_phred__float": "CADD_phred",
    "CADD_raw_hg19__float": "CADD_raw_hg19",
    "CADD_raw_rankscore_hg19__float": "CADD_raw_rankscore_hg19",
    "CADD_phred_hg19__float": "CADD_phred_hg19",
    # arup
    "BayesDel_noAF_score__mfloat": "BayesDel_noAF_score",
    "BayesDel_noAF_rankscore__mfloat": "BayesDel_noAF_rankscore",
    "BayesDel_noAF_pred__mstring": "BayesDel_noAF_pred",
    "BayesDel_addAF_pred__mstring": "BayesDel_addAF_pred",
    "BayesDel_addAF_rankscore__mfloat": "BayesDel_addAF_rankscore",
    "BayesDel_addAF_score__mfloat": "BayesDel_addAF_score",
    "AlphaMissense_score__mfloat": "AlphaMissense_score",
    "AlphaMissense_rankscore__mfloat": "AlphaMissense_rankscore",
    "AlphaMissense_pred__mstring": "AlphaMissense_pred",
}


def get_grch38_csra(x):
    # If there is a grch38 then we want to build the csra
    # using the following fields.
    # order of function composition is important here because
    # the next function in the chain
    # will alter the field names.
    if (grch38 := x.get("pos(1-based)", None)):
        x["grch38_csra"] = "|".join(["chr" + x.get("#chr"),
                                     grch38, x.get("ref"), x.get("alt"), "GRCh38"])
    return x


def clean_genes(ann):
    return ann.get("gene", "").split(";")


def none_to_nan(x):  # pragma: no mccabe
    """as per kms-5025, the null items were not getting passed to solr.
    for example, if we sent a list [1,2,3,None,5] then solr would only
    ingest [1,2,3,5]. This was corrupting our data since the index in the list
    act as place holders for the genetic and variant information. We fixed this
    by replacing None with "NaN" since solr will allow "NaN" to be used as
     a place holder for float fields. We use "None"
     if it is a string field and -999999 if its an int field.

     Any fields that have all Nones, -999999 or Nan get purged from the dict.
     """
    keys_marked_for_removal = []
    values_to_purge = (None, 'None', -999999, "NaN")
    for k, v in x.items():
        if isinstance(v, list) and k.endswith("__mstring"):
            if None in v:
                v = [val if val is not None else "None" for val in v]
                x[k] = v
                if all(val in values_to_purge for val in x[k]):
                    keys_marked_for_removal.append(k)
        if isinstance(v, list) and (
                k.endswith("__mfloat")):
            v = [val if val is not None else "NaN" for val in v]
            x[k] = v
            if all(val in values_to_purge for val in x[k]):
                keys_marked_for_removal.append(k)
        if isinstance(v, list) and (k.endswith("__mint")):
            v = [val if val is not None else -999999 for val in v]
            x[k] = v
            if all(val in values_to_purge for val in x[k]):
                keys_marked_for_removal.append(k)
    for key in keys_marked_for_removal:
        x.pop(key, None)
    return x


def remove_blanks_from_bayesdel_dataset(x):
    for k, v in x.items():
        if k.startswith('BayesDel') and '' in v:
            x[k] = [val for val in v if val != '']
    return x


register(
    input_type=DocType.TSV,
    output_type=DocType.DBNSFP,
    transformer=compose(
        filter_out_empty_chr_start,
        additional_clean_up,
        lambda x: assoc(x, "csra", kms.annotations.to_csra(x)),
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        lambda x: assoc(x, "gene", clean_genes(x)),
        remove_blanks_from_bayesdel_dataset,
        none_to_nan,
        lambda x: assoc(x, __TYPE__, DocType.DBNSFP.value),
        name_mapping(
            NAME_MAPPING, empty_values=(None, "."), split_chars=SPLIT_CHARS
        ),
        set_chr_start,
        get_grch38_csra,

    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.DBNSFP,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.GNOMAD.value)),
    is_header=True,
)



from cytoolz.curried import curry, dissoc
from . import Sink
import os

try:
    import ujson as json
except ImportError:
    import json


FILE_PATH = "/Users/ian/work/m2gen/markers/Fusion/"


@curry
class TempAlterationSink(Sink):
    def __init__(self, filename):
        pass

    def close(self):
        pass

    def write(self, unit):
        file_path = unit.get("file_path")
        rna_id = os.path.basename(file_path).split("_")[0]
        with open(
            os.path.join(FILE_PATH, f"{rna_id}_fusion.jsonl"),
            "a",
            encoding='utf-8'
        ) as fh:
            unit = dissoc(unit, "__type__", "file_path")
            fh.write(json.dumps(unit))
            fh.write("\n")



import json
import logging
import re
import sqlite3
from tempfile import NamedTemporaryFile

from cytoolz.curried import assoc, curry
from genomoncology.parse import __CHILD__, __TYPE__, DocType

from ..base import LazyFileSource

TAG_RE = re.compile(r"<[^>]+>")

_logger = logging.getLogger(__name__)


@curry
class MitomapFileSource(LazyFileSource):
    def __init__(self, dump_file_name, keep_db=False, **meta):
        self.meta = meta

        if __TYPE__ not in meta:
            self.meta = assoc(
                self.meta, __TYPE__, DocType.MITOMAP_RECORD.value
            )

        self.dump_file = open(dump_file_name, "r")
        self.table_names = [
            "mitomap.genbank",
            "mitomap.gbcontrol",
            "mitomap.genbank_count",
            "mitomap.genbank_haplogroup",
            "mitomap.mitotip",
            "mitomap.helix",
            "mitomap.apogee",
            "mitomap.polymorphism",
            "mitomap.hmtvar",
            "mitomap.mmutation",
            "mitomap.rtmutation"
        ]
        self.table_indexed_fields = {
            "mitomap.genbank": ["tpos", "qnt"],
            "mitomap.gbcontrol": ["tpos", "qnt"],
            "mitomap.genbank_count": ["haplogroup", "alt", "pos"],
            "mitomap.genbank_haplogroup": ["haplogroup"],
            "mitomap.mitotip": ["pos", "ref", "alt"],
            "mitomap.helix": ["position", "refna", "regna"],
            "mitomap.apogee": ["position", "refna", "regna"],
            "mitomap.polymorphism": ["position", "refna", "regna", "aachange"],
            "mitomap.hmtvar": ["pos", "ref", "alt"],
            "mitomap.mmutation": ["position", "refna", "regna"],
            "mitomap.rtmutation": ["position", "refna", "regna"]
        }
        self.lines = self.dump_file.readlines()
        self.table_line_boundaries = {}
        self.table_fields = {}
        # map of pos-ref-alt to set of haplogroups
        self.pra_map = {}
        # maps to gather counts during genbank table load
        self.var_haplo_cnt_map = {}
        self.haplo_cnt_map = {}
        self.fl_count_map = {}
        # set to get total fl count
        self.genbank_id_set = set()
        self.db = sqlite3.connect(":memory:")
        self.temp_file = NamedTemporaryFile(delete=not keep_db)
        if keep_db:
            _logger.debug("Mitomap database can be found at: %s", self.temp_file.name)
        self.db = sqlite3.connect(self.temp_file.name)
        self.cursor = self.db.cursor()
        self.create_tables_in_db()

    # METHODS TO GET INFO ON TABLE START/END LINES
    def line_num_end_of_input(self, line_number, lines):
        for the_line in lines:
            if the_line.startswith("\\."):
                return line_number
            line_number += 1

    def get_table_line_boundaries(self, lines, starting_num, table_name):
        end_of_input = self.line_num_end_of_input(starting_num, lines)
        self.table_line_boundaries[table_name] = (starting_num, end_of_input)
        return end_of_input

    # METHODS TO CREATE SQL STATEMENTS
    def format_fields_with_data_types(self, fields):
        structured_fields = []
        for i in range(len(fields)):
            structured_fields.append(fields[i].strip() + " TEXT")
        return ",".join(structured_fields)

    def create_question_mark_string(self, length_of_fields):
        question_marks = ["?"] * length_of_fields
        return "{}".format(",".join(question_marks))

    def create_index_for_table(self, table_name, indexed_fields):
        return "CREATE INDEX index_{0} ON {0} ({1});".format(
            table_name, ",".join(indexed_fields)
        )

    def create_table_statement(self, table_name, fields):
        structured_fields = self.format_fields_with_data_types(fields)
        return "CREATE TABLE {0}({1});".format(table_name, structured_fields)

    def insert_into_table_statement(self, table_name, fields):
        return "INSERT INTO {0}({1}) VALUES({2});".format(
            table_name,
            ",".join(fields),
            self.create_question_mark_string(len(fields)),
        )

    def select_data_statement(self, table_name, selected_fields):
        return "SELECT {0} FROM {1};".format(
            ",".join(selected_fields), table_name
        )

    def select_cr_count(self, tpos, qnt):
        return (
            "SELECT count(*) FROM gbcontrol WHERE tpos='{0}' AND "
            "qnt='{1}';".format(tpos, qnt)
        )

    def select_total_count(self, table_name):
        return "SELECT count(DISTINCT genbank_id) FROM {0}".format(table_name)

    def select_variant_haplogroup_count(self, haplogroup, qnt, tpos):
        return (
            "SELECT cnt FROM genbank_count WHERE haplogroup='{0}' "
            "and alt='{1}' and pos='{2}';".format(haplogroup, qnt, tpos)
        )

    def select_haplogroup_count(self, haplogroup):
        return (
            "SELECT count(genbank_id) FROM genbank_haplogroup WHERE "
            "haplogroup='{0}';".format(haplogroup)
        )

    def select_mitotip(self, tpos, tnt, qnt):
        return (
            "SELECT score FROM mitotip WHERE pos='{0}' "
            "AND ref='{1}' AND alt='{2}'".format(tpos, tnt, qnt)
        )

    def select_helix_info(self, tpos, tnt, qnt):
        return (
            "SELECT counts_hom, af_hom, gene FROM helix WHERE position='{0}' "
            "AND refna='{1}' AND regna='{2}'".format(tpos, tnt, qnt)
        )

    def select_apogee_info(self, tpos, tnt, qnt):
        return (
            "SELECT score, status FROM apogee WHERE position='{0}' "
            "AND refna='{1}' AND regna='{2}'".format(tpos, tnt, qnt)
        )

    def select_hmtvar_info(self, tpos, tnt, qnt):
        return (
            "SELECT disease_score, pathogenicity FROM hmtvar WHERE pos='{0}' "
            "AND ref='{1}' AND alt='{2}' "
            "AND NOT (disease_score = '\\N' AND pathogenicity = '\\N')"
        ).format(tpos, tnt, qnt)

    def select_mmutation_info(self, tpos, tnt, qnt):
        return (
            "SELECT status, dz FROM mmutation WHERE position='{0}' "
            "AND refna='{1}' AND regna='{2}'".format(tpos, tnt, qnt)
        )

    def select_rtmutation_info(self, tpos, tnt, qnt):
        return (
            "SELECT status, dz FROM rtmutation WHERE position='{0}'"
            "AND refna='{1}' AND regna='{2}'".format(tpos, tnt, qnt)
        )

    # METHODS TO EXECUTE SQL STATEMENTS
    def create_table(self, table_name, fields):
        self.cursor.execute(self.create_table_statement(table_name, fields))
        self.db.commit()

    def insert_into_table(
            self, table_name, fields, all_values, indexed_fields
    ):
        self.cursor.executemany(
            self.insert_into_table_statement(table_name, fields), all_values
        )
        if indexed_fields is not None:
            self.cursor.execute(
                self.create_index_for_table(table_name, indexed_fields)
            )
        self.db.commit()

    def get_table_data(self, table_name, selected_fields):
        self.cursor.execute(
            self.select_data_statement(table_name, selected_fields)
        )
        return self.cursor.fetchall()

    def get_fl_count(self, tpos, qnt):
        return self.fl_count_map[(tpos, qnt)]

    def get_cr_count(self, tpos, qnt):
        self.cursor.execute(self.select_cr_count(tpos, qnt))
        return self.cursor.fetchone()[0]

    def get_total_fl_count(self):
        return len(self.genbank_id_set)

    def get_total_count(self, table_name):
        self.cursor.execute(self.select_total_count(table_name))
        return self.cursor.fetchone()[0]

    def get_variant_haplogroup_count(self, haplogroup, qnt, tpos):
        self.cursor.execute(
            self.select_variant_haplogroup_count(haplogroup, qnt, tpos)
        )
        result = self.cursor.fetchone()
        if result is not None:
            return result[0]
        return 0  # return 0 if not present

    def get_haplogroup_count(self, haplogroup):
        self.cursor.execute(self.select_haplogroup_count(haplogroup))
        return self.cursor.fetchone()[0]

    def get_mitotip(self, tpos, tnt, qnt):
        self.cursor.execute(self.select_mitotip(tpos, tnt, qnt))
        result = self.cursor.fetchone()
        if result is not None:
            return result[0]
        return None  # return None if no score

    def get_helix_info(self, tpos, tnt, qnt):
        self.cursor.execute(self.select_helix_info(tpos, tnt, qnt))
        result = self.cursor.fetchone()
        if result is not None:
            return result[0], 195983, result[1], result[2]
        return None, None, None, None  # return None if no helix info

    def get_apogee(self, tpos, tnt, qnt):
        apogee_list = []
        self.cursor.execute(self.select_apogee_info(tpos, tnt, qnt))
        results = self.cursor.fetchall()
        for result in results:
            info = {"score": result[0], "consensus": result[1]}
            apogee_list.append(json.dumps(info))
        return apogee_list

    def get_hmtvar_mutations(self, tpos, tnt, qnt):
        hmtvar_list = []
        self.cursor.execute(self.select_hmtvar_info(tpos, tnt, qnt))
        results = self.cursor.fetchall()
        for result in results:
            info = {"disease_score": result[0] if result[0] != "\\N" else None,
                    "pathogenicity": result[1] if result[1] != "\\N" else None}
            hmtvar_list.append(json.dumps(info))
        return hmtvar_list

    def get_mmutation(self, tpos, tnt, qnt):
        mmutation_list = []
        self.cursor.execute(self.select_mmutation_info(tpos, tnt, qnt))
        results = self.cursor.fetchall()
        for result in results:
            info = {"status__mstring": result[0].replace(
                    "Cfrm", "Confirmed"), "dz": result[1]}
            mmutation_list.append(json.dumps(info))
        return mmutation_list

    def get_rtmutation(self, tpos, tnt, qnt):
        rtmutation_list = []
        self.cursor.execute(self.select_rtmutation_info(tpos, tnt, qnt))
        results = self.cursor.fetchall()
        for result in results:
            info = {"status__mstring": result[0].replace(
                    "Cfrm", "Confirmed"), "dz": result[1]}
            rtmutation_list.append(json.dumps(info))
        return rtmutation_list

    # METHODS TO CREATE AND POPULATE THE TABLES
    def parse_and_set_genbank_data(self, the_lines):
        data = []
        for line in the_lines:
            line = line.replace("\n", "")
            record = line.split("\t")
            # get the pos, ref, and alt to see if duplicate hgvs
            gb_fields = self.table_fields["mitomap.genbank"]
            pos = record[gb_fields.index("tpos")]
            ref = record[gb_fields.index("tnt")]
            alt = record[gb_fields.index("qnt")]
            # get haplogroup
            haplogroup = record[gb_fields.index("haplogroup")]
            pra_key = (pos, ref, alt)
            if pra_key not in self.pra_map.keys():
                self.pra_map[pra_key] = {"haplogroup_set": set([haplogroup])}
                # add the first occurrence to the genbank table
                data.append(tuple(record))
            else:
                self.pra_map[pra_key]["haplogroup_set"].add(haplogroup)

            # calculate fl_count
            fl_key = (pos, alt)
            if fl_key not in self.fl_count_map.keys():
                self.fl_count_map[fl_key] = 1
            else:
                self.fl_count_map[fl_key] = self.fl_count_map[fl_key] + 1

            # add genbank id to set to use to calculate total fl count
            genbank_id = record[gb_fields.index("genbank_id")]
            self.genbank_id_set.add(genbank_id)

        return data

    def set_table_data(self, the_lines):
        data = []
        for line in the_lines:
            line = line.replace("\n", "")
            record = line.split("\t")
            data.append(tuple(record))
        return data

    def create_and_populate_table(
            self, table_name, the_lines, fields, indexed_fields
    ):
        if table_name == "mitomap.genbank":
            data = self.parse_and_set_genbank_data(the_lines)
        else:
            data = self.set_table_data(the_lines)
        # remove mitomap. from the name when actually creating the db table
        table_name = table_name.replace("mitomap.", "")
        self.create_table(table_name, fields)
        self.insert_into_table(table_name, fields, data, indexed_fields)

    def create_tables_in_db(self):
        line_num = 0
        while line_num < len(self.lines):
            line = self.lines[line_num]
            if line.startswith("COPY") and line.split()[1] in self.table_names:
                table_name = line.split()[1]
                field_names = re.search(r"\((.*?)\)", line).group(1)
                self.table_fields[table_name] = list(
                    map(str.strip, field_names.split(","))
                )
                # get the line boundaries
                table_end_line = self.get_table_line_boundaries(
                    self.lines[line_num + 1:], line_num + 1, table_name
                )
                # create and populate the table
                self.create_and_populate_table(
                    table_name,
                    self.lines[line_num + 1: table_end_line],
                    self.table_fields[table_name],
                    self.table_indexed_fields[table_name],
                )
                line_num = table_end_line + 1
            line_num += 1

    def get_insertion(self, tnt, qnt):
        return qnt.replace(tnt, "", 1)

    def get_hgvs_g(self, tpos, tnt, qnt, ntchange):
        if ntchange == "insertion":
            insertion = self.get_insertion(tnt, qnt)
            return "NC_012920.1:g.{0}_{1}ins{2}".format(
                tpos, int(tpos) + 1, insertion
            )
        if ntchange == "deletion":
            return "NC_012920.1:g.{0}del{1}".format(tpos, tnt)
        return "NC_012920.1:g.{0}{1}>{2}".format(tpos, tnt, qnt)

    def get_csra(self, tpos, tnt, qnt):
        qnt = "-" if qnt == ":" else qnt
        tnt = "-" if tnt == ":" else tnt
        return f"chrMT|{tpos}|{tnt}|{qnt}|GRCh37"

    def get_freq(self, numer, denom):
        return 0 if denom == 0 else numer / denom

    def remove_html_tags(self, text):
        text_array = [
            non_tag_text.strip()
            for non_tag_text in TAG_RE.split(text)
            if non_tag_text.strip()
        ]
        return " ".join(text_array)

    def __iter__(self):  # pragma: no mccabe
        yield self.create_header()

        selected_fields = [
            "genbank_id",
            "tpos",
            "tnt",
            "qnt",
            "ntchange",
            "mmutid",
            "disease",
            "cal_aachange",
            "conservation",
            "haplogroup",
            "calc_locus",
        ]
        records = self.get_table_data("genbank", selected_fields)
        fl_total_count = self.get_total_fl_count()
        cr_total_count = self.get_total_count("gbcontrol")
        # used to store csra's and check for dups when building
        # polymorphism records.
        seen_records = set()

        for rec in records:
            # get fields for each record
            tpos = rec[selected_fields.index("tpos")]
            tnt = rec[selected_fields.index("tnt")]
            qnt = rec[selected_fields.index("qnt")]
            ntchange = rec[selected_fields.index("ntchange")]
            hgvs_g = self.get_hgvs_g(tpos, tnt, qnt, ntchange)
            csra = self.get_csra(tpos, tnt, qnt)
            disease = rec[selected_fields.index("disease")]
            aa_change = self.remove_html_tags(
                rec[selected_fields.index("cal_aachange")]
            )
            conservation = rec[selected_fields.index("conservation")]
            loci = rec[selected_fields.index("calc_locus")]
            # get haplogroups for this pos-ref-alt
            pra_key = (tpos, tnt, qnt)
            haplo_set = self.pra_map[pra_key]["haplogroup_set"]
            haplogroups = list(haplo_set)

            # get calculated fields
            fl_count = self.get_fl_count(tpos, qnt)
            cr_count = self.get_cr_count(tpos, qnt)
            fl_frequency = self.get_freq(fl_count, fl_total_count)
            cr_frequency = self.get_freq(cr_count, cr_total_count)
            mitoTIP = self.get_mitotip(tpos, tnt, qnt)
            h_count, h_total_count, h_frequency, h_gene = self.get_helix_info(
                tpos, tnt, qnt
            )
            apogee_list = self.get_apogee(tpos, tnt, qnt)
            hmtvar_list = self.get_hmtvar_mutations(tpos, tnt, qnt)
            mmutation_list = self.get_mmutation(tpos, tnt, qnt)
            rtmutation_list = self.get_rtmutation(tpos, tnt, qnt)

            var_haplo_counts = []
            haplo_counts = []
            for haplo in haplogroups:
                var_haplo_counts.append(
                    self.var_haplo_cnt_map.setdefault(
                        (haplo, qnt, tpos),
                        self.get_variant_haplogroup_count(haplo, qnt, tpos),
                    )
                )
                haplo_counts.append(
                    self.haplo_cnt_map.setdefault(
                        haplo, self.get_haplogroup_count(haplo)
                    )
                )
            seen_records.add(csra)
            yield {
                "hgvs_g": hgvs_g,
                "csra": csra,
                "disease": disease,
                "aa_change": aa_change,
                "conservation": conservation,
                "haplogroups": haplogroups,
                "loci": loci,
                "fl_count": fl_count,
                "cr_count": cr_count,
                "fl_frequency": fl_frequency,
                "cr_frequency": cr_frequency,
                "variant_haplogroup_counts": var_haplo_counts,
                "haplogroup_counts": haplo_counts,
                "mitoTIP": mitoTIP,
                "helix_count": h_count,
                "helix_total_count": h_total_count,
                "helix_frequency": h_frequency,
                "gene": h_gene,
                "apogee": apogee_list,
                "hmtvarmutation": hmtvar_list,
                "mmutation": mmutation_list,
                "rtmutation": rtmutation_list,
                "__type__": DocType.MITOMAP_RECORD.value,
            }

        # Parse mmutation records

        m_mutation_fields = [
            "position",
            "refna",
            "regna",
            "locus",
            "dz",
            "aa",
            "cons"
        ]
        m_mutation_records = self.get_table_data("mmutation", m_mutation_fields)
        for record in m_mutation_records:

            pos = record[m_mutation_fields.index("position")]
            refna = record[m_mutation_fields.index("refna")]
            regna = record[m_mutation_fields.index("regna")]
            gene = record[m_mutation_fields.index("locus")]
            disease = record[m_mutation_fields.index("dz")]
            aa_change = record[m_mutation_fields.index("aa")]
            conservation = record[m_mutation_fields.index("cons")]

            # If csra was already created in earlier records, do not display again.
            if (csra := self.get_csra(pos, refna, regna)) in seen_records:
                continue

            # If there are no results then assign an empty dict
            # which will get merged lower down and remove the key that has no value.
            if (hmtvarmutations := self.get_hmtvar_mutations(pos, refna, regna)) == []:
                hmtvarmutations = {}
            if (mmutations := self.get_mmutation(pos, refna, regna)) == []:
                mmutations = {}
            if (rtmutations := self.get_rtmutation(pos, refna, regna)) == []:
                rtmutations = {}
            seen_records.add(csra)
            yield {
                "csra": csra,
                "gene": gene,
                "disease": disease,
                "aa_change": aa_change,
                "conservation": conservation,
                # only add these fields if they have a value
                **({"hmtvarmutation": hmtvarmutations} if hmtvarmutations else {}),
                **({"mmutation": mmutations} if mmutations else {}),
                **({"rtmutation": rtmutations} if rtmutations else {}),
                "__type__": DocType.MITOMAP_RECORD.value,
            }
        # Parse rtmutation records

        rt_mutation_fields = [
            "position",
            "refna",
            "regna",
            "rna",
            "locus",
            "dz",
            "cons"
        ]
        rt_mutation_records = self.get_table_data("rtmutation", rt_mutation_fields)
        for record in rt_mutation_records:
            pos = record[rt_mutation_fields.index("position")]
            refna = record[rt_mutation_fields.index("refna")]
            regna = record[rt_mutation_fields.index("regna")]
            gene = record[rt_mutation_fields.index("locus")]
            disease = record[rt_mutation_fields.index("dz")]
            aa_change = record[rt_mutation_fields.index("rna")]
            conservation = record[rt_mutation_fields.index("cons")]

            # If csra was already created in earlier records, do not display again.
            if (csra := self.get_csra(pos, refna, regna)) in seen_records:
                continue
            # If there are no results then assign an empty dict
            # which will get merged lower down and remove the key that has no value.
            if (hmtvarmutations := self.get_hmtvar_mutations(pos, refna, regna)) == []:
                hmtvarmutations = {}
            if (mmutations := self.get_mmutation(pos, refna, regna)) == []:
                mmutations = {}
            if (rtmutations := self.get_rtmutation(pos, refna, regna)) == []:
                rtmutations = {}
            seen_records.add(csra)
            yield {
                "csra": csra,
                "gene": gene,
                "disease": disease,
                "aa_change": aa_change,
                "conservation": conservation,
                # only add these fields if they have a value
                **({"hmtvarmutation": hmtvarmutations} if hmtvarmutations else {}),
                **({"mmutation": mmutations} if mmutations else {}),
                **({"rtmutation": rtmutations} if rtmutations else {}),
                "__type__": DocType.MITOMAP_RECORD.value,
            }

        # Parse hmtvar records

        hmtvar_fields = [
            "pos",
            "ref",
            "alt",
            "locus"
        ]
        hmtvar_records = self.get_table_data("hmtvar", hmtvar_fields)
        for record in hmtvar_records:
            pos = record[hmtvar_fields.index("pos")]
            ref = record[hmtvar_fields.index("ref")]
            alt = record[hmtvar_fields.index("alt")]
            gene = record[hmtvar_fields.index("locus")]

            # If csra was already created in earlier records, do not display again.
            if (csra := self.get_csra(pos, ref, alt)) in seen_records:
                continue
            # If there are no results then assign an empty dict
            # which will get merged lower down and remove the key that has no value.
            if (hmtvarmutations := self.get_hmtvar_mutations(pos, refna, regna)) == []:
                hmtvarmutations = {}
            if (mmutations := self.get_mmutation(pos, refna, regna)) == []:
                mmutations = {}
            if (rtmutations := self.get_rtmutation(pos, refna, regna)) == []:
                rtmutations = {}
            seen_records.add(csra)
            yield {
                "csra": csra,
                "gene": gene,
                # only add these fields if they have a value
                **({"hmtvarmutation": hmtvarmutations} if hmtvarmutations else {}),
                **({"mmutation": mmutations} if mmutations else {}),
                **({"rtmutation": rtmutations} if rtmutations else {}),
                "__type__": DocType.MITOMAP_RECORD.value,
            }

        # Parse polymorphism records separately
        poly_fields = [
            "position",
            "refna",
            "regna",
            "aachange"
        ]
        polymorphism_records = self.get_table_data("polymorphism",
                                                   poly_fields)
        for poly in polymorphism_records:
            pos = poly[poly_fields.index("position")]
            refna = poly[poly_fields.index("refna")]
            regna = poly[poly_fields.index("regna")]

            # If the record was already seen then don't display it as
            # a polymorphism record.
            if (csra := self.get_csra(pos, refna, regna)) in seen_records:
                continue
            # If there are no results then assign an empty dict
            # which will get merged lower down and remove the key that has no value.
            if (hmtvarmutations := self.get_hmtvar_mutations(pos, refna, regna)) == []:
                hmtvarmutations = {}
            if (mmutations := self.get_mmutation(pos, refna, regna)) == []:
                mmutations = {}
            if (rtmutations := self.get_rtmutation(pos, refna, regna)) == []:
                rtmutations = {}

            yield {
                "csra": csra,
                "hgvs_g": self.get_csra(pos, refna, regna),
                "aa_change": poly[poly_fields.index("aachange")],
                # only add these fields if they have a value
                **({"hmtvarmutation": hmtvarmutations} if hmtvarmutations else {}),
                **({"mmutation": mmutations} if mmutations else {}),
                **({"rtmutation": rtmutations} if rtmutations else {}),
                "__type__": DocType.MITOMAP_RECORD.value,
            }

        self.cursor.close()
        self.db.close()

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "meta": self.meta,
            "fields": [
                "hgvs_g",
                "csra",
                "disease",
                "aa_change",
                "conservation",
                "haplogroups",
                "loci",
                "fl_count",
                "cr_count",
                "fl_frequency",
                "cr_frequency",
                "variant_haplogroup_counts",
                "haplogroup_counts",
                "mitoTIP",
                "helix_count",
                "helix_total_count",
                "helix_frequency",
                "gene",
                "apogee",
                "hmtvarmutation",
                "mmutation",
                "rtmutation"
            ],
        }



from .base import Source, JsonlFileSource, TextFileSource, NullSource
from .bed import BedFileSource
from .delimited import DelimitedFileSource
from .paths import CollectFilePathsSource
from .excel import ExcelSource
from .aggregated import (
    AggregatedCOSMICSources,
    AggregatedFileSource,
    AggregatedGenieSNVFileSource,
    AggregatedGenieCNVFileSource,
    AggregatedGenieCTXFileSource,
    AggregatedOmimFileSource,
    AggregatedCOSMICNonSNVSources,
    AggregatedGenieGenes
)
from .maf import MAFSource
from .xml import XMLSource
from . import one_off_sources
from .annotations_filter import AnnotationsFilterFileSource

__all__ = (
    "Source",
    "BedFileSource",
    "JsonlFileSource",
    "TextFileSource",
    "NullSource",
    "DelimitedFileSource",
    "CollectFilePathsSource",
    "ExcelSource",
    "AggregatedCOSMICSources",
    "AggregatedCOSMICNonSNVSources",
    "AggregatedFileSource",
    "AggregatedGenieSNVFileSource",
    "AggregatedGenieCNVFileSource",
    "AggregatedGenieCTXFileSource",
    "MAFSource",
    "XMLSource",
    "AnnotationsFilterFileSource",
    "AggregatedOmimFileSource",
    "AggregatedGenieGenes",
)


def get_one_off_source(source_name):
    return one_off_sources.source_name_map.get(source_name)



import re
import logging
from functools import partial

from cytoolz.curried import curry, assoc

from genomoncology.parse import (
    is_call_or_variant,
    __TYPE__,
    DocType,
    __CHILD__,
)
from genomoncology.parse.ensures import ensure_collection
import gosdk

from genomoncology.cli.const import GRCH37


ANNOTATIONS = [
    "GNOMAD__AF__mfloat",
    "ExAC__AF__float",
    "canonical_alteration",
    "canonical_c_dot",
    "canonical_gene",
    "canonical_mutation_type",
    "canonical_p_dot",
    "clinvar__CLNSIG__string",
    "display",
    "mutation_type",
    "representations",
    "hgvs_g",
    "alteration",
    "gene_annotations",
]
EXCLUDE_BUILD_FIELDS = ["is_gene_annotation"]
RE_REL_ORIEN_SAME = re.compile(
    r'''
        (?: \](chr)?([0-9]+|[xyXY]):[0-9]+\]\w
        |   \w\[(chr)?([0-9]+|[xyXY]):[0-9]+\[
        )
    ''',
    re.VERBOSE
)
RE_REL_ORIEN_REVERSE = re.compile(
    r'''
        (?: \[(chr)?([0-9]+|[xyXY]):[0-9]+\[\w
        |   \w\](chr)?([0-9]+|[xyXY]):[0-9]+\]
        )
    ''',
    re.VERBOSE
)

_logger = logging.getLogger(__name__)


@curry
async def hotspot_annotate(data, data_set, data_set_version):
    annotated_hotspots = annotate_hotspots(data)

    annotations_list = await gosdk.async_call_with_retry(
        gosdk.sdk.annotations.load_annotations,
        data_set=data_set,
        data_set_version=data_set_version,
        data=annotated_hotspots,
    )

    return annotations_list


@curry
async def annotate_genes(
    data, delete_if_exists=False, build=GRCH37
):
    genes = [record['gene'] for record in data]
    genes_annotations_list = await gosdk.async_call_with_retry(
        gosdk.sdk.annotations.post_annotations_match,
        batch=genes,
        delete_if_exists=delete_if_exists,
        build=build,
    )
    return (genes_annotations_list[gene] for gene in genes)


@curry
async def annotate_match(
    data,
    delete_if_exists=False,
    build=GRCH37,
    annotation_bundle_version=None,
    hgvs_batch=False,
):
    csra_batch = get_csra_batch_or_raw_batch(data, hgvs_batch, build)

    if csra_batch:
        annotations_list = await gosdk.async_call_with_retry(
            gosdk.sdk.annotations.post_annotations_match,
            batch=csra_batch,
            delete_if_exists=delete_if_exists,
            build=build,
        )
    else:
        annotations_list = None

    annotated_calls = annotate_match_calls(
        annotations_list, data, annotation_bundle_version, hgvs_batch
    )

    _logger.debug(
        ("annotate_match: call_count=%s first_call=%s csra_batch=%s annotated_count=%s"
            " first_annotated=%s"),
        len(data),
        data[:1],
        csra_batch[:4] if csra_batch else None,
        len(annotated_calls),
        annotated_calls[:1]
    )

    return annotated_calls


@curry
async def normalize_genie_hash(batch, delete_if_exists=False, build=GRCH37):
    # This function is only ever used for the GENIE aggregators
    # and is only used for aggregating SNVs for normalization
    # Just calls match and returns the normalization dict
    batch = list(batch)
    b = [r["variant"] for r in batch]
    response = await gosdk.async_call_with_retry(
        gosdk.sdk.annotations.post_generate_hgvs_g,
        batch=b,
        build=build,
    )
    for record in batch:
        variant = record.pop("variant")
        record["hgvs_g"] = response["normalized"].get(variant, variant)
    return batch


@curry
def annotate_sv_match(
    calls,
    delete_if_exists=False,
    build=GRCH37,
    region_match_tolerance=0.5,
    cnv_match_tolerance=0.33,
):
    SV_MUTATION_TYPES = ["INS", "INV", "DEL", "DUP", "CNV", "BND"]
    batch = []
    seen_bnd_ids = []
    for record in calls:
        if record["__type__"] != "HEADER":
            if record.get("mutation_type") not in SV_MUTATION_TYPES:
                _logger.error("unknown SVTYPE: %s", record.get('mutation_type'))
            if record["mutation_type"] == "BND":
                if record.get("mate_id"):
                    if record["id"] not in seen_bnd_ids:
                        batch.append(parse_bnd_sv_hashes(record, calls, True))
                        seen_bnd_ids.append(record["id"])
                        seen_bnd_ids.append(record["mate_id"][0])
                else:
                    batch.append(parse_bnd_sv_hashes(record, calls, False))

            else:
                batch.append(record["sv_hash"])

    if not batch:
        return []

    annotations = gosdk.call_with_retry(
        gosdk.sdk.annotations.post_annotations_sv_match,
        batch=batch,
        delete_if_exists=delete_if_exists,
        build=build,
        region_match_tolerance=region_match_tolerance,
        cnv_match_tolerance=cnv_match_tolerance,
    )
    return annotations.get("results", [])


@curry
async def load_annotations(
        data, data_set=None, data_set_version=None, build=GRCH37,
        is_custom=False):
    # do not add the build field (or set it equal to None)
    # if certain fields are present
    if is_custom:
        endpoint = gosdk.sdk.annotations.postapiannotationsload_custom
    else:
        endpoint = gosdk.sdk.annotations.load_annotations
    if len(data) > 0 and any(
            [key in EXCLUDE_BUILD_FIELDS for key in data[0].keys()]
    ):
        build = None
    response = await gosdk.async_call_with_retry(
        endpoint,
        data=override_hgvs([{**d, "build": build} for d in data]),
        data_set=data_set,
        data_set_version=data_set_version,
    )

    _logger.debug(
        "load_annotations: call_count=%s, first_record=%s",
        len(data),
        data[:1]
    )

    # response of warehouse is blank, returning data for downstream processing
    return data, response


def override_hgvs(data):
    data = ensure_collection(data)
    return list(map(up_build, data))


def get_csra_batch_or_raw_batch(data, hgvs_batch, build):
    # Check if input data is just a list of hashes or genes
    if hgvs_batch or all(isinstance(d, str) for d in data):
        return list(data)
    # Otherwise input data is a list of VARIANT_CALL or some other datatype
    else:
        return convert_to_csra([{**d, "build": build} for d in data])


def convert_to_csra(data, add_build=True):
    # create batch of CSRA strings for calling HGVS API
    data = ensure_collection(data)
    calls_only = filter(is_call_or_variant, data)
    csra_batch = list(
        to_csra(call, add_build=add_build) for call in calls_only
    )
    return csra_batch


def get_mutation_type(record):
    if "*" in record["Residue"]:
        return ["Nonstop Extension"]
    elif "X" in record["Residue"]:
        return ["Splice Donor Site", "Splice Acceptor Site"]
    elif "*" in record["Variants"] and "|" in record["Variants"]:
        return ["Substitution - Missense", "Substitution - Nonsense"]
    elif "*" in record["Variants"]:
        return ["Substitution - Nonsense"]
    else:
        return ["Substitution - Missense"]


def annotate_hotspots(data):
    print(data)
    annotations = []
    num_pattern = re.compile("[0-9]+")
    alpha_pattern = re.compile("[a-zA-Z]+")
    for record in data:
        ann = {}
        ann["gene"] = record["Gene"]
        ann["is_hotspot_annotation"] = True
        ann["cancerhotspots_count__int"] = record["Samples"]
        ann["cancerhotspots_tumor_type_composition__mstring"] = (
            record["Tumor Type Composition"].replace(":", "=").split("|")
        )
        if record["Type"] == "single residue":
            ref_aa = num_pattern.sub("", record["Residue"])
            pos = alpha_pattern.sub("", record["Residue"])
            if ref_aa != "X":
                ann["codon_start"] = pos
                ann["codon_end"] = pos
                ann["ref_aa_start"] = ref_aa
                ann["ref_aa_end"] = ref_aa
            else:
                ann["nearest_codon"] = pos
            ann["hotspot_mutation_type"] = get_mutation_type(record)
        elif record["Type"] == "in-frame indel":
            start_end = record["Residue"].split("-")
            if len(start_end) > 1:
                ann["codon_start"] = start_end[0]
                ann["codon_end"] = start_end[1]
            else:
                ann["codon_start"] = start_end[0]
                ann["codon_end"] = start_end[0]
            ann["hotspot_mutation_type"] = [
                "Insertion - In frame",
                "Deletion - In frame",
                "Complex - insertion inframe",
                "Complex - deletion inframe",
                "Complex - compound substitution",
            ]

        annotations.append(ann)

    return annotations


def annotate_match_calls(
    annotations_list, data, annotation_bundle_version, hgvs_batch
):
    if hgvs_batch:
        func = partial(
            add_annotation_hgvs_match,
            annotations_list,
            annotation_bundle_version,
        )
    else:
        func = partial(
            add_annotation_match, annotations_list, annotation_bundle_version
        )
    annotated_calls = list(map(func, data))
    return annotated_calls


# This method is used to format the API result if the input is a batch of
# raw hgvs_g/hgvs_c variant strings instead of CALL or VARIANT doctypes
def add_annotation_hgvs_match(
    annotations_list, expected_bundle_version, call: str
):
    annotations = {}
    if annotations_list:
        annotations = (
            annotations_list[call]
        )
        if expected_bundle_version is not None:

            actual_bundle_version = (
                annotations.get("annotation_bundle_version")
                or expected_bundle_version
            )

            if expected_bundle_version != actual_bundle_version:
                msg = (
                    "Annotations Process Terminated due to "
                    "annotations_bundle_version "
                    "changing during a run. \n"
                    "Expected annotation bundle version: {expected} \n"
                    "Actual annotation bundle version: {actual}".format(
                        expected=expected_bundle_version,
                        actual=actual_bundle_version,
                    )
                )
                raise AnnBundleVersionAssertionError(msg)
    return annotations


def add_annotation_match(
    annotations_list,
        expected_bundle_version,
        call: dict
):  # pragma: no mccabe

    if annotations_list and is_call_or_variant(call):
        csra = to_csra(call)
        annotations = (
            annotations_list[csra]
        )

        # if the annotation_bundle does not match, raise an exception
        if expected_bundle_version is not None:

            actual_bundle_version = (
                annotations.get("annotation_bundle_version")
                or expected_bundle_version
            )

            if expected_bundle_version != actual_bundle_version:
                msg = (
                    "Annotations Process Terminated due to "
                    "annotations_bundle_version "
                    "changing during a run. \n"
                    "Expected annotation bundle version: {expected} \n"
                    "Actual annotation bundle version: {actual}".format(
                        expected=expected_bundle_version,
                        actual=actual_bundle_version,
                    )
                )
                raise AnnBundleVersionAssertionError(msg)

        call["annotations"] = annotations
        call[__TYPE__] = f"ANNOTATED_MATCH_{call.get(__TYPE__, 'CALL')}"

    elif DocType.HEADER.is_a(call):
        call[__CHILD__] = f"ANNOTATED_MATCH_{call.get(__CHILD__, 'CALL')}"

    return call


def annotate_calls(
    annotations_list,
    data,
    annotation_bundle_version,
    filter_anns=None,
    hgvs_batch=False,
):
    # If we're not working with CALL objects but rather strings
    if hgvs_batch:
        func = partial(
            add_hgvs_annotation,
            annotations_list,
            annotation_bundle_version,
            filter_anns=filter_anns,
        )
    else:
        func = partial(
            add_annotation,
            annotations_list,
            annotation_bundle_version,
            filter_anns=filter_anns,
        )
    annotated_calls = list(map(func, data))
    return annotated_calls


def add_annotation(
    annotations_list,
    expected_bundle_version,
    call: dict,
    filter_anns=None,
):  # pragma: no mccabe

    if annotations_list and is_call_or_variant(call):
        csra = to_csra(call)
        annotations = (
            annotations_list[csra]
        )

        # if the annotation_bundle does not match, raise an exception
        if expected_bundle_version is not None:

            actual_bundle_version = (
                annotations.get("annotation_bundle_version")
                or expected_bundle_version
            )

            if expected_bundle_version != actual_bundle_version:
                msg = (
                    "Annotations Process Terminated due to "
                    "annotations_bundle_version "
                    "changing during a run. \n"
                    "Expected annotation bundle version: {expected} \n"
                    "Actual annotation bundle version: {actual}".format(
                        expected=expected_bundle_version,
                        actual=actual_bundle_version,
                    )
                )
                raise AnnBundleVersionAssertionError(msg)

        # filter the annotations if there was a filter file passed in
        if filter_anns is not None and isinstance(filter_anns, list):
            annotations = handle_filtering_annotations(
                annotations, filter_anns
            )

        call["annotations"] = annotations
        call[__TYPE__] = f"ANNOTATED_{call.get(__TYPE__, 'CALL')}"

    elif DocType.HEADER.is_a(call):
        call[__CHILD__] = f"ANNOTATED_{call.get(__CHILD__, 'CALL')}"

    return call


def add_hgvs_annotation(
    annotations_list,
    expected_bundle_version,
    call: str,
    filter_anns=None,
):

    annotations = {}
    if annotations_list:
        annotations = (
            annotations_list[call]
        )
        if expected_bundle_version is not None:

            actual_bundle_version = (
                annotations.get("annotation_bundle_version")
                or expected_bundle_version
            )

            if expected_bundle_version != actual_bundle_version:
                msg = (
                    "Annotations Process Terminated due to "
                    "annotations_bundle_version "
                    "changing during a run. \n"
                    "Expected annotation bundle version: {expected} \n"
                    "Actual annotation bundle version: {actual}".format(
                        expected=expected_bundle_version,
                        actual=actual_bundle_version,
                    )
                )
                raise AnnBundleVersionAssertionError(msg)

        if filter_anns is not None and isinstance(filter_anns, list):
            annotations = handle_filtering_annotations(
                annotations, filter_anns
            )
    return annotations


def handle_filtering_annotations(annotations, filter_anns):
    gene_annotations = []
    for gene_ann in annotations.get("gene_annotations", []):
        gene_ann_filtered = {
            k: v
            for k, v in gene_ann.items()
            if k in filter_anns or k == "gene"
        }
        if len(gene_ann_filtered) > 0:
            gene_annotations.append(gene_ann_filtered)
    annotations["gene_annotations"] = gene_annotations
    annotations = {
        k: v
        for k, v in annotations.items()
        if k in filter_anns or k == "gene_annotations"
    }
    return annotations


def up_build(call: dict):
    hgvs_g = call.get("hgvs_g")
    build = call.get("build")
    has_both = hgvs_g is not None and build is not None

    if has_both and not hgvs_g.endswith(f"|{build}"):
        hgvs_parts = hgvs_g.split("|")
        if len(hgvs_parts) >= 4:
            hgvs_parts = hgvs_parts[:4] + [build]
            call = assoc(call, "hgvs_g", "|".join(hgvs_parts))

    return call


def to_csra(call: dict, add_build=True, add_chr=True):
    chromosome = f"chr{call.get('chr')}" if add_chr else call.get("chr")
    csra_data = (
        chromosome,
        str(call.get("start")),
        call.get("ref") or "-",
        call.get("alt") or "-",
        call.get("build") or GRCH37,
    )
    return "|".join(csra_data[0:] if add_build else csra_data[0:4])


def format_chromosome_start_end(data):
    if data is None:
        return None, None, None
    data = [None if not str(x).strip() else x for x in data]
    return data[0], data[1], data[2]


def get_annotation_bundle_version():
    annotation_bundle_version = gosdk.call_with_retry(
        gosdk.sdk.info.get_annotation_bundle_version
    )
    return annotation_bundle_version.annotation_bundle_version


def parse_bnd_sv_hashes(record, data, has_mate_id):
    if has_mate_id:
        for item in data:
            if item.get("id", None) == record["mate_id"][0]:
                chr = record["chr_info"]
                fusion = item["chr_info"]
                extra_info = record["extra_info"]
                rel_orien = determine_relative_orientation(record["alt"])

                record["sv_hash"] = f"{chr}|{fusion}|{extra_info}{rel_orien}"
                item["sv_hash"] = f"{chr}|{fusion}|{extra_info}{rel_orien}"
                return f"{chr}|{fusion}|{extra_info}{rel_orien}"
    else:
        # Alternative parsing if no mate_id present
        chr = record["chr_info"]
        fusion = f"{record.get('missing_mate_fusion_chr')}:" \
                 f"{record.get('missing_mate_fusion_pos')}"
        extra_info = record["extra_info"]
        rel_orien = determine_relative_orientation(record["alt"])
        record["sv_hash"] = f"{chr}|{fusion}|{extra_info}{rel_orien}"
        return f"{chr}|{fusion}|{extra_info}{rel_orien}"


def determine_relative_orientation(alt):
    if RE_REL_ORIEN_SAME.match(alt):
        return "|REL_ORIEN=same"
    if RE_REL_ORIEN_REVERSE.match(alt):
        return "|REL_ORIEN=reverse"
    return ""


class AnnBundleVersionAssertionError(AssertionError):
    pass



from cytoolz.curried import assoc, compose

from genomoncology.parse.doctypes import DocType, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping
from genomoncology.pipeline.transformers.functions import filter_keys

NAME_MAPPING = {
    # info
    "clinvar__CLNSIG__mstring": "clinvar__CLNSIG__mstring",
    "clinvar__variant_ID__string": "clinvar__variant_ID__string",
    "clinvar__CLNREVSTAT__mstring": "clinvar__CLNREVSTAT__mstring",
    "clinvar__CLNVI__mstring": "clinvar__CLNVI__mstring",
    "clinvar__CLNDN__mstring": "clinvar__CLNDN__mstring",
    "clinvar__all_submission_interpretations__mstring": "clinvar__all_submission_interpretations__mstring",  # noqa: E501
    "clinvar__all_submission_review_statuses__mstring": "clinvar__all_submission_review_statuses__mstring",  # noqa: E501
    "clinvar__all_submission_conditions__mstring": "clinvar__all_submission_conditions__mstring",  # noqa: E501
    "clinvar__all_submission_submitter__mstring": "clinvar__all_submission_submitter__mstring",  # noqa: E501
    "clinvar__all_submission_origin__mstring": "clinvar__all_submission_origin__mstring",  # noqa: E501
    "clinvar_alteration__string": "clinvar_alteration__string",
    "hgvs_g": "hgvs_g",
}

register(
    input_type=DocType.CLINVAR_XML,
    output_type=DocType.CLINVAR_XML,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.CLINVAR_XML.value),
        filter_keys(
            {
                "clinvar__CLNSIG__mstring",
                "clinvar__variant_ID__string",
                "clinvar__CLNREVSTAT__mstring",
                "clinvar__CLNVI__mstring",
                "clinvar__CLNDN__mstring",
                "clinvar__all_submission_interpretations__mstring",
                "clinvar__all_submission_review_statuses__mstring",
                "clinvar__all_submission_conditions__mstring",
                "clinvar__all_submission_submitter__mstring",
                "clinvar__all_submission_origin__mstring",
                "clinvar_alteration__string",
                "hgvs_g",
            }
        ),
        name_mapping(NAME_MAPPING),
    ),
)



from cytoolz.curried import curry
from glom import glom

from genomoncology.parse import DocType
from genomoncology.parse.ensures import ensure_collection
from .base import FileSink
from ..converters import to_str


@curry
class TsvFileSink(FileSink):
    def __init__(
        self, filename, columns=None, include_header=False, delimiter="\t"
    ):
        super().__init__(filename)
        self.delimiter = delimiter
        self.columns = ensure_collection(columns)
        self.include_header = include_header

    def write(self, unit):
        if self.include_header or DocType.HEADER.is_not(unit):
            super(TsvFileSink.func, self).write(unit)

    def convert(self, unit):
        self.columns = self.columns or DocType.get_default_columns(unit)

        if DocType.HEADER.is_a(unit):
            data = DocType.get_header_names(self.columns)
        else:
            data = [to_str(glom(unit, g, default="")) for g in self.columns]

        return self.delimiter.join(data)



from cytoolz.curried import assoc, compose

from genomoncology import kms
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


SPLIT_CHARS = ";"


NAME_MAPPING = {
    "chr": "chromosome",
    "start": "position",
    "ref": "ref",
    "alt": "alt",
    "pathology__mstring": "pathology",
    "tRNA__string": "tRNA",
    "tRNA_structural_domain__string": "tRNA_structural_domain",
}


def clean_pathology(value):
    return [path.strip() for path in value]


register(
    input_type=DocType.TSV,
    output_type=DocType.MITO_TRNA,
    transformer=compose(
        lambda x: assoc(
            x,
            "pathology__mstring",
            clean_pathology(x.get("pathology__mstring", "")),
        ),
        lambda x: assoc(x, "csra", kms.annotations.to_csra(x, add_chr=False)),
        lambda x: assoc(
            x, "hgvs_g", kms.annotations.to_csra(x, add_chr=False)
        ),
        lambda x: assoc(x, __TYPE__, DocType.MITO_TRNA.value),
        name_mapping(
            NAME_MAPPING, empty_values=(None, "."), split_chars=SPLIT_CHARS
        ),
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.MITO_TRNA,
    transformer=compose(
        lambda x: assoc(x, __CHILD__, DocType.MITO_TRNA.value)
    ),
    is_header=True,
)



import asyncio
import logging
from multiprocessing import Process, Queue

from bravado_core.exception import SwaggerError
from cytoolz.curried import curry, dissoc
from genomoncology.parse import DocType, ensures
from genomoncology.pipeline.transformers.functions import filter_private_keys

import gosdk

from . import Sink

_logger = logging.getLogger(__name__)

EOL = "EOL"


async def make_call(data_type, in_queue):
    sdk_module = getattr(gosdk.sdk, f"warehouse_{data_type}")
    sdk_func = getattr(sdk_module, f"create_warehouse_{data_type}")

    unit = in_queue.get()

    while unit != EOL:
        try:
            await gosdk.async_call_with_retry(sdk_func, data=unit)
        except (SwaggerError, OSError, ConnectionError, TimeoutError) as e:
            _logger.exception("failed to load %s.%s: %s", sdk_module, sdk_func, e)

        unit = in_queue.get()


def background_runner(data_type, queue):
    gosdk.setup_sdk()
    asyncio.run(make_call(data_type, queue))


class LoadWarehouseSink(Sink):

    DATA_TYPE = None

    def __init__(self, _, state, num_workers=10):
        assert self.DATA_TYPE, "Do not instantiate LoadWarehouseSink directly."

        self.queue = Queue(maxsize=1000)
        self.workers = []

        self.num_workers = num_workers
        self.build = state.build

        for _ in range(num_workers):
            worker = Process(
                target=background_runner,
                args=(state, self.DATA_TYPE, self.queue),
            )
            worker.start()
            self.workers.append(worker)

    def close(self):
        for _ in range(self.num_workers):
            self.queue.put(EOL)

        for worker in self.workers:
            worker.join()

    def write(self, unit):
        unit = ensures.ensure_collection(unit)
        unit = filter(DocType.HEADER.is_not, unit)
        unit = map(filter_private_keys, unit)
        unit = [dissoc(d, "annotations") for d in unit]
        unit = [{**d, "build": self.build} for d in unit]

        self.queue.put(unit)

        return unit


@curry
class LoadWarehouseVariantsSink(LoadWarehouseSink):

    DATA_TYPE = "variants"


@curry
class LoadWarehouseFeaturesSink(LoadWarehouseSink):

    DATA_TYPE = "features"



import logging

import asyncio
from cytoolz.curried import curry, partition_all, concat, map

_logger = logging.getLogger(__name__)


def async_processor(
    state, async_func, **kw
):
    return [
        map(async_func(**kw)),
        # roll up the futures in batches, await, flatten
        partition_all(state.parallel_batches),
        map(do_await(state)),
        concat,
    ]


@curry
def do_await(state, pending):
    """
    pending must be a batch of futures from the async processor
    """
    loop = state.runner.get_loop()

    results = loop.run_until_complete(
        asyncio.gather(*pending, return_exceptions=True)
    )

    for result in results:
        if isinstance(result, Exception):
            raise result if state.hard_failure else _logger.error("%s", result)
        else:
            yield result



from cytoolz.curried import assoc, compose

from genomoncology import kms
from genomoncology.parse.doctypes import DocType, __TYPE__, __CHILD__
from genomoncology.pipeline.transformers.functions import filter_keys
from genomoncology.pipeline.transformers.registry import register
from genomoncology.pipeline.transformers.mapper_classes import VariantMapper


def clean_PROT__string(value):
    if isinstance(value, str):
        value = value.replace("%3D", "=").replace("%3A", ":")
    return value


def clean_DNA__string(value):
    if isinstance(value, str):
        value = value.replace("%3D", "=").replace("%3A", ":")
    return value


def clean_PHEN_string(value):
    if isinstance(value, tuple):
        value = ",".join(value)
    return value[1:-1].replace("%2C", ",").replace("_", " ")


register(
    input_type=DocType.CALL,
    output_type=DocType.HGMD,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.HGMD.value),
        filter_keys(
            {
                "CLASS__string",
                "DB__string",
                "DNA__string",
                "GENE__string",
                "PHEN__string",
                "PROT__string",
                "ID__string",
                "hgvs_g",
                "csra",
            }
        ),
        lambda x: assoc(
            x, "PROT__string", clean_PROT__string(x.get("PROT__string"))
        ),
        lambda x: assoc(
            x, "DNA__string", clean_PROT__string(x.get("DNA__string"))
        ),
        # clean PHEN: remove lead/trail quotes, replace underscores w/ spaces
        lambda x: assoc(
            x, "PHEN__string", clean_PHEN_string(x.get("PHEN__string", ""))
        ),
        lambda x: assoc(x, "ID__string", x.get("id", None)),
        lambda x: assoc(x, "csra", kms.annotations.to_csra(x)),
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        # convert Call to Variant
        VariantMapper(),
    ),
    is_header=False,
)

register(
    input_type=DocType.CALL,
    output_type=DocType.HGMD,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.HGMD.value)),
    is_header=True,
)



import re
from typing import Union
import logging
import copy

from cytoolz.curried import curry, assoc, keyfilter, get_in
from glom import glom
from glom.core import PathAccessError

from genomoncology.parse.doctypes import DocType

_logger = logging.getLogger(__name__)


@curry
def transform(glom_spec: Union[dict, str], value: dict):
    # e.g.: transform(call, {"AN_EAS": "annotations.GNOMAD__AN_EAS__mint.0"})
    if value:
        try:
            transformed = glom(value, glom_spec)
            return transformed
        except PathAccessError as e:
            _logger.warning("transform path error: %s", e)
            return None


@curry
def filter_keys(whitelist, value):
    return keyfilter(lambda k: k in whitelist, value)


@curry
def get_in_field(field_name, value):
    field_name = field_name.split(".")
    return get_in(field_name, value)


@curry
def rename(old_name, new_name, value):
    return assoc(value, new_name, get_in_field(old_name, value))


@curry
def filter_private_keys(value):
    return keyfilter(lambda k: not k.startswith("__"), value)


MULTI_FIELD = re.compile("^.*__m(string|float|int)$")
DELS = re.compile("[|:;,]")


def get_split_regex(split_chars):
    if split_chars is None:
        return DELS
    return re.compile(f"[{split_chars}]")


def dot_to_none(v):
    if v != ".":
        return v


def split_value(field_value, split_chars=None):
    if isinstance(field_value, str):
        field_value = [
            dot_to_none(v)
            for v in get_split_regex(split_chars).split(field_value)
        ]
    return field_value


def is_multi_field(field_name):
    return MULTI_FIELD.fullmatch(field_name)


def get_boolean_value(s):
    return s.lower() == "true"


@curry
def name_mapping(
        mapping, value, empty_values=(None,), check_multi=True, split_chars=None,
        ignore_period=False
):
    new_dict = {}
    for (new_name, old_name) in mapping.items():
        if ignore_period:
            # in some files we do not want to split on the "."
            field_value = get_in([old_name], value)
        else:
            field_value = get_in_field(old_name, value)

        # add __record__
        if old_name == "__record__":
            new_dict = assoc(new_dict, new_name, field_value)
            continue

        if field_value not in empty_values:
            if check_multi and is_multi_field(new_name):
                field_value = split_value(field_value, split_chars=split_chars)

            new_dict = assoc(new_dict, new_name, field_value)

    return new_dict


@curry
def add_flag(bed_filter, flag_name, value):
    if not DocType.HEADER.is_a(value):
        item = (value.get("chr"), value.get("start"))

        bed_value_list = []
        is_present = False
        for i in bed_filter.search(item):
            bed_value_list += i.data
            is_present = True

        if bed_value_list:
            value = assoc(value, f"{flag_name}__mstring", bed_value_list)
        value = assoc(value, f"{flag_name}__boolean", is_present)

    return value


@curry
def add_rollup(value, key="__rollup__"):
    if len(list(value)) == 1 and DocType.HEADER.is_a(value[0]):
        return value
    call_list = []
    for call in value:
        call_list.append(call)
        if call.get("zyg", "") == "heterozygous" and any(
            [
                allele_index == 0
                for allele_index in call.get("allele_indices", [])
            ]
        ):
            # create the ref call for the rollup
            ref_call = copy.deepcopy(call)
            ref_call["alt"] = ref_call["ref"]
            call_list.append(ref_call)
    rollup = list(map(filter_private_keys, call_list))
    return map(lambda child: assoc(child, key, rollup), value)


def turn_gene_to_list(x):
    gene = x.get("gene")
    return [gene] if gene else []



from .clinvar_xml import ClinvarXMLSource
from .mitomap import MitomapFileSource
from .uniprot import UniprotFileSource
from .mtdb import MTDBFileSource

source_name_map = {
    "clinvar_xml": ClinvarXMLSource,
    "mitomap": MitomapFileSource,
    "uniprot": UniprotFileSource,
    "mtdb": MTDBFileSource,
}

__all__ = (
    "ClinvarXMLSource",
    "MitomapFileSource",
    "UniprotFileSource",
    "MTDBFileSource",
)



from . import comparators, converters, sinks, sources, filters
from .runner import run_pipeline


__all__ = (
    "comparators",
    "converters",
    "sinks",
    "sources",
    "filters",
    "run_pipeline",
)



from cytoolz.curried import filter, map, compose, concat

from genomoncology.parse.ensures import ensure_collection
from genomoncology.pipeline.converters import non_null
from genomoncology.pipeline.transformers import get_in_field, transform


def calls_to_alterations(calls):
    calls = ensure_collection(calls)
    pipeline = compose(
        concat, filter(non_null), map(get_in_field("annotations.alteration"))
    )
    alterations = list(pipeline(calls))
    return alterations


def vie_results_to_alts(vies):
    vies = ensure_collection(vies)

    # using the VIE response, loop through each variant and add it to either
    # the alterations or benign_alterations lists, depending on the
    # "match_field" value
    alterations = []
    benign_alterations = []

    # now, let's loop through each item and get the vie info
    for vie_result in vies:
        vie_data = vie_result.get("variant_interpretations", [])
        alt_name = vie_data[0].get("alteration")
        match_field = vie_data[0].get("match_field")
        if match_field == "alteration":
            alterations.append(alt_name)
        elif match_field == "benign_alteration":
            benign_alterations.append(alt_name)
    return alterations, benign_alterations


def transform_match_response(response, default_spec):
    transformer = transform(default_spec)
    return map(transformer, response)



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import (
    register,
    name_mapping,
)

SPLIT_CHARS = ",;"

inheritance_map = {
    "AD": "Autosomal Dominant",
    "AR": "Autosomal Recessive",
    "MT": "Mitochondrial",
    "SD": "Semidominant",
    "Undetermined": "Undetermined",
    "XL": "X-Linked"
}

NAME_MAPPING = {
    "gene": "GENE SYMBOL",
    "hgnc_id": "GENE ID (HGNC)",
    "disease__string": "DISEASE LABEL",
    "disease_id__string": "DISEASE ID (MONDO)",
    "mode_of_inheritance": "MOI",
    "classification__string": "CLASSIFICATION",
    "url__string": "ONLINE REPORT"

}


def clean_hgnc(x):
    return x["hgnc_id"].replace("HGNC:", "")


def get_mode_of_inheritance(x):
    return inheritance_map.get(x["mode_of_inheritance"])


register(
    input_type=DocType.TSV,
    output_type=DocType.CLINGEN_GENE_DISEASE,
    transformer=compose(
        lambda x: assoc(x, "hgnc_id", clean_hgnc(x)),
        lambda x: assoc(x, __TYPE__, DocType.CLINGEN_GENE_DISEASE.value),
        lambda x: assoc(x, "mode_of_inheritance", get_mode_of_inheritance(x)),
        lambda x: assoc(x, "is_gene_annotation", True),
        name_mapping(NAME_MAPPING, empty_values=(None, "", ".", [])),
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.CLINGEN_GENE_DISEASE,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.CLINGEN_GENE_DISEASE.value)
    ),
    is_header=True,
)



INPUT_IS_FILE_PATH_LIST = "__INPUT_IS_FILE_PATH_LIST__"
OUTPUT_IS_EXCEL_FILE = "__OUTPUT_IS_EXCEL_FILE__"
SKIP_CONVERT_TO_JSON = "__SKIP_CONVERT_TO_JSON__"

GRCH37 = "GRCh37"
GRCH38 = "GRCh38"
BUILDS = [GRCH37, GRCH38]



import logging

from cytoolz.curried import curry
from genomoncology.cli.const import GRCH37
from genomoncology.parse.doctypes import DocType
from genomoncology.parse.ensures import ensure_collection

import gosdk

_logger = logging.getLogger(__name__)


@curry
def boundaries(data, build=GRCH37):
    gene_list = gosdk.call_with_retry(
        gosdk.sdk.genes.gene_boundaries,
        name=ensure_collection(data),
        build=build
    )
    if gene_list.not_found:
        _logger.warning('genes not found: %s', gene_list.not_found)

    def _to_result(gene):
        return {
            '__type__': DocType.GENE.value,
            'chromosome': gene['chromosome'],
            'start': gene['start'],
            'end': gene['end'],
            'gene': gene['gene'],
        }

    return map(_to_result, gene_list.results)



import logging

from cytoolz.curried import curry

import gosdk

_logger = logging.getLogger(__name__)


@curry
def process_transcript_batch(data, genes_only=False):
    batch = [
        "%s|%s|%s" % (record["chromosome"], record["start"], record["end"])
        for record in data
    ]
    _logger.debug("get_transcripts_batch: batch=%s", batch)

    results = gosdk.call_with_retry(
        gosdk.sdk.region_search.region_search_batch,
        batch=batch
    )

    if genes_only:
        genes = set()
        for result in results['results']:
            for transcript in result['transcripts']:
                genes.add(transcript['gene'])
        return genes

    return results['results']
